// 005 u_c_test_case_list_3
// 24 jan 2009

// -- (C) Felix John COLIBRI 2009
// -- documentation: http://www.felix-colibri.com

// -- c_test_method
// --   c_test_case
// --     c_test_case_list

(*$r+*)

unit u_c_test_case_list_3;
  interface
    uses Classes
        , SysUtils // Exceptions
        , u_c_basic_object;

    type c_failure= Class(Exception);

         t_po_test_method= Procedure of Object;

         c_test_method_ref= // one "test_method"
             Class(c_basic_object)
               // -- m_name: an optional method name
               m_po_test_method: t_po_test_method;
               m_enabled: Boolean;

               Constructor create_test_method_ref(p_name: String;
                   p_po_test_method: t_po_test_method);
               function f_c_self: c_test_method_ref;
               function f_display_test_method_ref: String;
            end; // c_test_method_ref

         c_test_case_list= Class; // forward

         c_test_case= // one "test_case"
             Class(c_basic_object)
               Public
                 // -- m_name:
                 m_c_test_method_ref_list: tStringList;
                 // -- to be able to fire the form events
                 m_c_parent_test_case_list: c_test_case_list;
                 m_enabled: Boolean;

                 Constructor create_test_case(p_name: String;
                     p_c_parent_test_case_list: c_test_case_list);

                 function f_c_self: c_test_case;

                 function f_display_test_case: String;
                 procedure display_test_method_list;

                 function f_test_method_ref_count: Integer;
                 function f_c_test_method(p_test_method_index: Integer): c_test_method_ref;
                 function f_index_of(p_test_method_name: String): Integer;
                 function f_c_find_test_method_ref_by_name(p_test_method_name: String): c_test_method_ref;
                 function f_c_find_test_method_ref(p_c_test_method_ref: c_test_method_ref): c_test_method_ref;

                   procedure add_test_method(p_method_name: String; p_c_test_method: t_po_test_method);
                 procedure add_test_methods; Virtual; // Abstract;
                 procedure add_published_methods(p_c_class_ref: tClass);

                 procedure check(p_condition: Boolean; p_text: String);

                 // procedure Setup; Virtual;
                 procedure run_test_case;
                 // procedure Teardown; Virtual;

                 Destructor Destroy; Override;

               protected
                 procedure Setup; Virtual;
                 procedure Teardown; Virtual;
             end; // c_test_case

         c_test_case_class_ref= Class of c_test_case;
         t_po_after_test_method_event=
             Procedure(p_c_test_method: c_test_method_ref; p_success: Boolean;
                 p_error_message: String) of Object;
         t_po_before_test_case_event=
             Procedure(p_c_test_case: c_test_case) Of Object;
         t_po_after_test_case_event= Procedure(p_c_test_case: c_test_case;
             p_run_count, p_error_count: Integer) of Object;

         c_test_case_list= // "test_case" list
             Class(c_basic_object)
               m_c_test_case_list: tStringList;

               // -- mainly for gui display
               m_after_test_method: t_po_after_test_method_event;
               m_before_test_case: t_po_before_test_case_event;
               m_after_test_case: t_po_after_test_case_event;

               Constructor create_test_case_list(p_name: String);

               function f_c_self: c_test_case_list;
               function f_test_case_count: Integer;
               function f_c_test_case(p_test_case_index: Integer): c_test_case;
               function f_index_of(p_test_case_name: String): Integer;
               function f_c_find_by_test_case(p_test_case_name: String): c_test_case;
               procedure add_test_case(p_test_case_name: String; p_c_test_case: c_test_case);
               procedure display_test_case_list;

               procedure add_test_case_class(p_c_test_case_ref: c_test_case_class_ref);
               procedure run_test_case_list;

               Destructor Destroy; Override;
             end; // c_test_case_list

    function f_c_test_case_list: c_test_case_list;
    procedure set_test_case_list_trace(p_do_trace: Boolean);

  implementation
    uses u_c_display, u_display_hex_2
        ;

    var s_do_trace: Boolean= False;

    procedure set_test_case_list_trace(p_do_trace: Boolean);
      begin
        s_do_trace:= p_do_trace;
      end; // set_test_case_list_trace

    procedure display_trace(p_text: String);
      begin
        if s_do_trace
          then display(p_text);
      end; // display_trace

    // -- c_test_method_ref

    Constructor c_test_method_ref.create_test_method_ref(p_name: String;
        p_po_test_method: t_po_test_method);
      begin
        Inherited create_basic_object(p_name);

        m_po_test_method:= p_po_test_method;
        m_enabled:= True;
      end; // create_test_method_ref

    function c_test_method_ref.f_c_self: c_test_method_ref;
      begin
        Result:= Self;
      end; // f_c_self

    function c_test_method_ref.f_display_test_method_ref: String;
      begin
        with tMethod(m_po_test_method) do
          Result:= Format('%-40s  code %4x  data %4x  Self %4x',
              [m_name, Integer(Code), Integer(Data), Integer(Self)]);
      end; // f_display_test_method_ref

    // -- c_test_case

    Constructor c_test_case.create_test_case(p_name: String;
        p_c_parent_test_case_list: c_test_case_list);
      begin
        Inherited create_basic_object(p_name);
        m_c_parent_test_case_list:= p_c_parent_test_case_list;

        m_c_test_method_ref_list:= tStringList.Create;

        // -- will call the descendent method which will add all its test_methods
        add_test_methods;

        m_enabled:= True;
      end; // create_test_case

    function c_test_case.f_c_self: c_test_case;
      begin
        Result:= Self;
      end; // f_c_self
        
    function c_test_case.f_display_test_case: String;
      begin
        Result:= Format('%-15s', [m_name]);
      end; // f_display_test_case

    function c_test_case.f_test_method_ref_count: Integer;
      begin
        Result:= m_c_test_method_ref_list.Count;
      end; // f_test_method_ref_count

    function c_test_case.f_c_test_method(p_test_method_index: Integer): c_test_method_ref;
      begin
        Result:= c_test_method_ref(m_c_test_method_ref_list.Objects[p_test_method_index]);
      end; //  f_c_test_method

    function c_test_case.f_index_of(p_test_method_name: String): Integer;
      begin
        Result:= m_c_test_method_ref_list.IndexOf(p_test_method_name);
      end; // f_index_of

    function c_test_case.f_c_find_test_method_ref_by_name(p_test_method_name: String): c_test_method_ref;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_test_method_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_test_method_ref(m_c_test_method_ref_list.Objects[l_index_of]);
      end; // f_c_find_test_method_ref_by_name

    function c_test_case.f_c_find_test_method_ref(p_c_test_method_ref: c_test_method_ref): c_test_method_ref;
      var l_test_method_ref_index: Integer;
      begin
        display_trace('> c_test_case.f_c_find_test_method_ref '+ p_c_test_method_ref.m_name
            + ' '+ f_integer_to_hex(Integer(tMethod(p_c_test_method_ref.m_po_test_method).Code)));

        // -- first search by name
        Result:= f_c_find_test_method_ref_by_name(p_c_test_method_ref.m_name);

        if Result<> Nil
          then display_trace('found_by_name ');

        if Result= Nil
          then
            for l_test_method_ref_index:= 0 to f_test_method_ref_count- 1 do
            begin
              display_trace('addr '+ f_integer_to_hex(Integer(tMethod(f_c_test_method(l_test_method_ref_index).m_po_test_method).Code)));
              if tMethod(f_c_test_method(l_test_method_ref_index).m_po_test_method).Code
                  = tMethod(p_c_test_method_ref.m_po_test_method).Code
                then begin
                    Result:= f_c_test_method(l_test_method_ref_index);
                    Break;
                  end;
            end;

        display_trace('< c_test_case.f_c_find_test_method_ref '+ f_display_TF(Result<> Nil));
      end; // f_c_find_test_method_ref

    procedure c_test_case.add_test_method(p_method_name: String; p_c_test_method: t_po_test_method);
      var l_c_test_method: c_test_method_ref;
      begin
        display_trace('add_a_test_method method_name '+ p_method_name); // + p_c_test_method.Name);
        l_c_test_method:= c_test_method_ref.create_test_method_ref(p_method_name, p_c_test_method);
        if f_c_find_test_method_ref(l_c_test_method)= Nil
          then m_c_test_method_ref_list.AddObject('', l_c_test_method)
          else l_c_test_method.Free;
      end; // add_test_method

    procedure c_test_case.add_test_methods;
        // -- keep in case descendent do not overwrite this method (manual test method addition)
      begin
      end; // add_test_methods

    procedure c_test_case.add_published_methods(p_c_class_ref: tClass);

      procedure _add_test_method(p_method_name: ShortString);
        var l_pt_method_address: Pointer;
            l_method_code_and_data: TMethod;

            l_c_test_method_ref: c_test_method_ref;
            // l_po_test_method: t_po_test_method;
        begin
          display_trace(p_method_name);

          // -- only test under some condition
          // -- in our case, check that
          // --  - the Published method was not yet manually added
          // --  - the method name contains "test" somewhere

          if (m_c_test_method_ref_list.IndexOf(p_method_name)>= 0)
              or (Pos('test', LowerCase(p_method_name))<= 0)
            then Exit;

          // -- get the method address from its name
          l_pt_method_address:= MethodAddress(p_method_name);
          if not (l_pt_method_address= nil)
            then begin
                // -- create the c_method_test, with a NIL method pointer for now
                l_c_test_method_ref:= c_test_method_ref.create_test_method_ref(p_method_name, Nil);

                // -- create a (code+data) method record.
                // --  the code is this method
                l_method_code_and_data.Code:= l_pt_method_address;
                // --  the data is the c_test_case (NOT the c_test_method_ref)
                l_method_code_and_data.Data:= Self;

                // -- initialize the method pointer
                l_c_test_method_ref.m_po_test_method:= t_po_test_method(l_method_code_and_data);

                // -- add this test method to the list
                if f_c_find_test_method_ref(l_c_test_method_ref)= Nil
                  then m_c_test_method_ref_list.AddObject(p_method_name, l_c_test_method_ref)
                  else l_c_test_method_ref.Free;
              end
            else display_bug_stop('no_methodaddress_for '+ p_method_name);
        end; // _add_test_method

      type t_method_table= packed record
                             m_vmt_method_count: SmallInt;
                             // -- the methods name array
                           end;
      var l_pt_method_table: ^t_method_table;
          l_pt_name: ^ShortString;
          l_method_name_index: Integer;
          // , l_method_name_in_table_index: Integer

      begin // add_published_methods
        // -- uses information aboutn System.MethodName
        asm
          mov EAX, [p_c_class_ref]
          // -- get the pointer to the method table
          mov EAX, [EAX].vmtMethodTable
          mov[l_pt_method_table], EAX
        end;

        if l_pt_method_table<> nil
          then begin
              l_pt_name:= Pointer(PChar(l_pt_method_table)+ 8);

              // -- analyze the Vmt
              for l_method_name_index:= 1 to l_pt_method_table.m_vmt_method_count do
              begin
                _add_test_method(l_pt_name^);

                // -- next name
                l_pt_name:= Pointer(PChar(l_pt_name)+ length(l_pt_name^)+ 7)
              end; // for l_method_name_index
            end; // if l_pt_method_table<> nil
      end; // add_published_methods

    procedure c_test_case.display_test_method_list;
      var l_test_method_ref_index: Integer;
      begin
        display_trace(m_name+ ' ['+ IntToStr(f_test_method_ref_count)+ ']');

        for l_test_method_ref_index:= 0 to f_test_method_ref_count- 1 do
          display_trace(f_c_test_method(l_test_method_ref_index).f_display_test_method_ref);
      end; // display_test_method_list

    procedure c_test_case.setup;
        // -- still there in case descendent does not override
      begin
      end; // setup

    procedure c_test_case.check(p_condition: Boolean; p_text: String);
      begin
        if not p_condition
          then Raise c_failure.Create(p_text);
      end; // check

    procedure c_test_case.run_test_case;
      var l_c_test_method: c_test_method_ref;
          l_error_message: String;
          l_error_count: Integer;
          
      function f_setup: Boolean;
        begin
          display_trace('f_setup');
          Result:= True;
          Try
            SetUp;
          Except
            On e: Exception do
              begin
                l_error_message:= 'setup_exception '+ e.Message;
                Result:= False;
              end;
          end; // try except
        end; // f_setup

      function f_run: Boolean;
        begin
          display_trace('f_run');
          Result:= True;
          Try
            // -- call the test method
            l_c_test_method.m_po_test_method;
          except
            // -- do nothing to avoid an exception dialog
            // -- the l_success will notify the result to the callerbe
            on e : c_failure do
              begin
                Inc(l_error_count);
                Result:= False;
                l_error_message:= 'failure '+ e.Message;
              end;
            on e : Exception do
              begin
                Inc(l_error_count);
                Result:= False;
                l_error_message:= 'exception  '+ e.Message;
              end;
          end; // try except
        end; // f_run

      function f_teardown: Boolean;
        begin
          display_trace('f_teardown');
          Result:= True;
          Try
            Teardown;
          Except
            On e: Exception do
              begin
                l_error_message:= 'teardown_exception '+ e.Message;
                Result:= False;
              end;
          end; // try except
        end; // f_teardown

      var l_test_method_ref_index: Integer;
          l_success: Boolean;
          l_run_count: Integer;
          // -- debug
          l_save_indentation: Integer;

      begin
        display_trace('> run_test_case '+ m_name+ ' '+ IntToStr(f_test_method_ref_count));

        with m_c_parent_test_case_list do
          if Assigned(m_before_test_case)
            then m_before_test_case(Self);

        l_run_count:= 0; l_error_count:= 0;
        for l_test_method_ref_index:= 0 to f_test_method_ref_count- 1 do
        begin
          l_save_indentation:= f_indentation;
          l_error_message:= '';

          l_c_test_method:= f_c_test_method(l_test_method_ref_index);
          with l_c_test_method do
            if m_enabled
              then begin
                  Inc(l_run_count);

{
                  setup;
                  l_success:= False;

                  Try
                    try
                      // -- call the test method
                      m_po_test_method;
                      l_success:= True;
                    except
                      // -- do nothing to avoid an exception dialog
                      // -- the l_success will notify the result to the callerbe
                      on e : c_failure do
                        begin
                          Inc(l_error_count);
                          l_error_message:= 'failure '+ e.Message;
                        end;
                      on e : Exception do
                        begin
                          Inc(l_error_count);
                          l_error_message:= 'exception  '+ e.Message;
                        end;
                    end; // try except
                  Finally
                    with m_c_parent_test_case_list do
                      if Assigned(m_after_test_method)
                        then m_after_test_method(l_c_test_method, l_success, l_error_message);

                    tear_down;
                  end; // try finally
}
                  // -- avoid "uninitialized" warning
                  l_success:= True;

                  Try
                    l_success:= f_setup and f_run and f_teardown;
                  Finally
                    with m_c_parent_test_case_list do
                      if Assigned(m_after_test_method)
                        then m_after_test_method(l_c_test_method, l_success, l_error_message);
                  end;
                end; // with l_c_test_method

          set_indentation(l_save_indentation);
        end; // for l_test_method_ref_index

        // -- notify whold c_test_case result
        with m_c_parent_test_case_list do
          if Assigned(m_after_test_case)
            then m_after_test_case(Self, l_run_count, l_error_count);

        display_trace('< run_test_case ');
      end; // run_test_case

    procedure c_test_case.Teardown;
        // -- still there in case descendent does not override
      begin
      end; // Teardown

    Destructor c_test_case.Destroy;
      var l_test_method_ref_index: Integer;
      begin
        for l_test_method_ref_index:= 0 to f_test_method_ref_count- 1 do
          f_c_test_method(l_test_method_ref_index).Free;
        m_c_test_method_ref_list.Free;

        InHerited;
      end; // Destroy

    // -- c_test_case_list

    Constructor c_test_case_list.create_test_case_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_c_test_case_list:= tStringList.Create;
      end; // create_test_case_line

    function c_test_case_list.f_c_self: c_test_case_list;
      begin
        Result:= Self;
      end; // f_c_self

    function c_test_case_list.f_test_case_count: Integer;
      begin
        Result:= m_c_test_case_list.Count;
      end; // f_test_case_count

    function c_test_case_list.f_c_test_case(p_test_case_index: Integer): c_test_case;
      begin
        Result:= c_test_case(m_c_test_case_list.Objects[p_test_case_index]);
      end; //  f_c_test_case

    function c_test_case_list.f_index_of(p_test_case_name: String): Integer;
      begin
        Result:= m_c_test_case_list.IndexOf(p_test_case_name);
      end; // f_index_of

    function c_test_case_list.f_c_find_by_test_case(p_test_case_name: String): c_test_case;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_test_case_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_test_case(m_c_test_case_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_test_case_list.add_test_case(p_test_case_name: String; p_c_test_case: c_test_case);
      begin
        m_c_test_case_list.AddObject(p_test_case_name, p_c_test_case);
      end; // add_test_case

    procedure c_test_case_list.add_test_case_class(p_c_test_case_ref: c_test_case_class_ref);
      var l_c_test_case: c_test_case;
          l_class_name: String;
      begin
        // -- create the test case from his class ref
        l_c_test_case:= p_c_test_case_ref.create_test_case('', Self);
        l_class_name:= l_c_test_case.ClassName;
        display_trace(l_class_name);
        add_test_case(l_class_name, l_c_test_case);

        // -- add all the Published methods, spelunking into the Vmt
        l_c_test_case.add_published_methods(p_c_test_case_ref);

        display_trace('> display_test_method_list_of_'+ l_class_name+ ' '+ l_c_test_case.m_name);
        l_c_test_case.display_test_method_list;
        display_trace('< display_test_method_list');
      end; // add_test_case_class

    procedure c_test_case_list.display_test_case_list;
      var l_test_case_index: Integer;
      begin
        display_trace(m_name+ ' '+ IntToStr(f_test_case_count));

        for l_test_case_index:= 0 to f_test_case_count- 1 do
          display_trace(f_c_test_case(l_test_case_index).f_display_test_case);
      end; // display_test_case_list

    procedure c_test_case_list.run_test_case_list;
      var l_test_case_index: Integer;
      begin
        for l_test_case_index:= 0 to f_test_case_count- 1 do
          with f_c_test_case(l_test_case_index) do
            if m_enabled
              then run_test_case;
      end; // run_test_case_list

    Destructor c_test_case_list.Destroy;
      var l_test_case_index: Integer;
      begin
        for l_test_case_index:= 0 to f_test_case_count- 1 do
          f_c_test_case(l_test_case_index).Free;
        m_c_test_case_list.Free;

        Inherited;
      end; // Destroy

    // -- singleton

    var g_c_test_case_list: c_test_case_list= Nil;
    
    function f_c_test_case_list: c_test_case_list;
      begin
        if g_c_test_case_list= Nil
          then g_c_test_case_list:= c_test_case_list.create_test_case_list('');
        Result:= g_c_test_case_list;
      end; // f_c_test_case_list

    begin // u_c_test_case_list
    end. // u_c_test_case_list
