// 005 u_test_gui
// 24 jan 2009

// -- (C) Felix John COLIBRI 2009
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_test_gui;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls
       , ImgList
       , u_c_test_case_list_3
       ;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    exit_: TButton;
    clear_: TButton;
    run_test_case_list_: TButton;
    Splitter1: TSplitter;
    PageControl1: TPageControl;
    display_: TTabSheet;
    Memo1: TMemo;
    framework_: TTabSheet;
    Splitter2: TSplitter;
    TreeView1: TTreeView;
    test_result_memo_: TMemo;
    state_imagelist: TImageList;
    run_imagelist: TImageList;
    procedure FormCreate(Sender: TObject);
    procedure close_Click(Sender: TObject);
    procedure clear_Click(Sender: TObject);
    procedure run_test_case_list_Click(Sender: TObject);
    procedure TreeView1Click(Sender: TObject);
  private
    m_display_test_case_name: String;

    function f_c_treenode_from_test_name(p_case_or_method_name: String): tTreeNode;
    procedure handle_after_test_method(p_c_test_method: c_test_method_ref;
        p_success: Boolean; p_error_message: String);
    procedure handle_before_test_case(p_c_test_case: c_test_case);
    procedure handle_after_test_case(p_c_test_case: c_test_case;
        p_run_count, p_error_count: Integer);

  public
    // -- for automatic run
    procedure initialize_gui;
  end; // tForm1

    var Form1: TForm1;

    procedure initialize_test_gui;

  implementation
    uses u_c_log, u_c_display
       , u_exe

       ;

    {$R *.DFM}

    const // -- image index for TreeView run : tTreeNode.ImageIndex
          k_no_run_image= 0;
          k_running_image= 1;
          k_failure_image= 2;
          k_success_image= 3;

          // -- image index for TreeView state  : tTreeNode.StateIndex
          k_selected_treenode= 1;
          k_unselected_treenode= 2;

    procedure TForm1.FormCreate(Sender: TObject);
      begin
        initialize_display(Memo1.Lines);
        initialize_default_log;
      end; // FormCreate

    procedure TForm1.close_Click(Sender: TObject);
      begin
        Close;
      end; // close_Click

    procedure TForm1.clear_Click(Sender: TObject);
      begin
        clear_display;
      end; // clear_Click

    // -- gui

    function TForm1.f_c_treenode_from_test_name(p_case_or_method_name: String): tTreeNode;
      var l_tree_index: Integer;
      begin
        Result:= Nil;

        with TreeView1, Items do
          for l_tree_index:= 0 to Count- 1 do
            if Items[l_tree_index].Text= p_case_or_method_name
              then begin
                  Result:= Items[l_tree_index];
                  Break;
                end;

        if Result<> Nil
          then display('f_c_treenode_from_test_name '+ Result.Text)
          else display('f_c_treenode_from_test_name NIL');
      end; // f_c_treenode_from_test_name

    procedure TForm1.TreeView1Click(Sender: TObject);
        // -- toggle individual nodes to enable / disapble this c_test_case
        // --   or c_test_mathod
      var l_c_selected_node: TTreeNode;
      begin
        l_c_selected_node:= TreeView1.Selected;

        if l_c_selected_node<> Nil
          then
            if c_test_case(l_c_selected_node.Data) is c_test_case
              then
                with c_test_case(l_c_selected_node.Data) do
                begin
                  m_enabled:= not m_enabled;
                  if m_enabled
                    then l_c_selected_node.StateIndex:= k_selected_treenode
                    else l_c_selected_node.StateIndex:= k_unselected_treenode;
                end
              else
                if c_test_method_ref(l_c_selected_node.Data) is c_test_method_ref
                  then
                    with c_test_method_ref(l_c_selected_node.Data) do
                    begin
                      m_enabled:= not m_enabled;
                      if m_enabled
                        then l_c_selected_node.StateIndex:= k_selected_treenode
                        else l_c_selected_node.StateIndex:= k_unselected_treenode;
                    end;

        TreeView1.Repaint;
      end; // TreeView1Click

    // -- test case

    procedure TForm1.handle_before_test_case(p_c_test_case: c_test_case);
      begin
        m_display_test_case_name:= p_c_test_case.m_name;
      end; // handle_befor_test_case

    procedure TForm1.handle_after_test_method(p_c_test_method: c_test_method_ref;
        p_success: Boolean; p_error_message: String);
        // -- called after each c_test_method_ref.run
      var l_c_tree_node: tTreeNode;
      begin
        l_c_tree_node:= f_c_treenode_from_test_name(p_c_test_method.m_name);

        if p_success
          then begin
              display('  test_ok '+ p_c_test_method.m_name);

              if l_c_tree_node<> Nil
                then l_c_tree_node.ImageIndex:= k_success_image;
            end
          else begin
              display('  *** test_error '+ p_c_test_method.m_name+ ' '+ p_error_message);

              // -- display test case only once
              if m_display_test_case_name<> ''
                then begin
                    test_result_memo_.Lines.Add(m_display_test_case_name);
                    m_display_test_case_name:= '';
                  end;
              test_result_memo_.Lines.Add('  '+ p_c_test_method.m_name+ ' ['+ p_error_message+ ']');

              if l_c_tree_node<> Nil
                then begin
                    l_c_tree_node.ImageIndex:= k_failure_image;
                    // -- also toggle the c_test_case image
                    l_c_tree_node.Parent.ImageIndex:= k_failure_image;
                  end;
            end;
        TreeView1.Repaint;
      end; // handle_after_test_method

    procedure TForm1.handle_after_test_case(p_c_test_case: c_test_case;
        p_run_count, p_error_count: Integer);
      var l_c_treenode: tTreeNode;
      begin
        l_c_treenode:= f_c_treenode_from_test_name(p_c_test_case.m_name);

        if (p_error_count= 0) and (l_c_treenode<> Nil)
          then l_c_treenode.ImageIndex:= k_success_image
          else l_c_treenode.ImageIndex:= k_failure_image;
      end; // handle_after_test_case

    procedure TForm1.initialize_gui;

      procedure build_treeview;

        procedure reset_node(p_c_treenode: tTreeNode);
          begin
            p_c_treenode.ImageIndex:= k_no_run_image;
            p_c_treenode.StateIndex:= k_selected_treenode;
          end; // reset_node

        var l_test_case_index: Integer;
            l_c_test_case_treenode: tTreeNode;
            l_test_method_index: Integer;

        begin // build_treeview
          with f_c_test_case_list, TreeView1.Items do
          begin
            BeginUpdate;

            Clear;

            for l_test_case_index:= 0 to f_test_case_count- 1 do
              with f_c_test_case(l_test_case_index) do
              begin
                // -- add this at the level 0
                l_c_test_case_treenode:= AddObject(Nil, m_name, f_c_self);
                reset_node(l_c_test_case_treenode);

                for l_test_method_index:=0 to f_test_method_ref_count- 1 do
                  with f_c_test_method(l_test_method_index) do
                    reset_node(AddChildObject(l_c_test_case_treenode, m_name, f_c_self));
              end;

            TreeView1.FullExpand;
            EndUpdate;
          end; // with g_c_test_case_list, TreeView1.Items
        end; // build_treeview

      begin // initialize_gui
        with f_c_test_case_list do
        begin
          build_treeview;
          m_after_test_method:= handle_after_test_method;
          m_before_test_case:= handle_before_test_case;
          m_after_test_case:= handle_after_test_case;
        end; // with g_c_test_case_list
      end; // initialize_gui

    procedure TForm1.run_test_case_list_Click(Sender: TObject);
        // -- run all the checked c_method_test, and gather statistics
      var l_treenode_index: integer;
      begin
        display_line;

        // -- reset_treeview_state
        for l_treenode_index:= 0 to TreeView1.Items.Count- 1 do
          TreeView1.Items[l_treenode_index].ImageIndex:= k_no_run_image;

        f_c_test_case_list.run_test_case_list;
      end; // run_test_case_list_Click

    procedure initialize_test_gui;
      begin
        Application.Initialize;
        Application.CreateForm(TForm1, Form1);
        Form1.initialize_gui;
        Application.Run;
      end; // initialize_test_gui

    end.


